from dataclasses import dataclass
from px_settings.contrib.django import settings as s


__all__ = 'Settings', 'settings',


@s('pxd_legate')
@dataclass
class Settings:
    ADMIN_REGISTER_CONTENT_TYPE: bool = True
    ADMIN_REGISTER_PERMISSION: bool = True


settings = Settings()



from typing import Any, Callable, List, Optional, Sequence, Tuple
from django.contrib.contenttypes.models import ContentType
from django.db import models

from wcd_device_recognizer.utils import model_bulk_get_or_create
from pxd_combinable_groups.utils import get_group_model

from ..models import ObjectAccess
from ..utils import upset
from . import gatherer


Group = get_group_model()
ObjectDefinition = Tuple[ContentType, int]


def _obj_to_raw(objects: Sequence[models.Model]):
    cache = ContentType.objects.get_for_models(*{x.__class__ for x in objects})

    return [(cache[x.__class__], x.pk) for x in objects]


def _should(*items: Sequence[Optional[Sequence[Any]]]):
    return [item is not None and len(item) != 0 for item in items]


def change_access_raw(
    user_id: int,
    object_definitions: Sequence[ObjectDefinition],
    committer: Callable[[Sequence[int], Optional[Sequence[int]]], List[int]],
    permission_ids: Optional[Sequence[int]] = None,
    group_ids: Optional[Sequence[int]] = None,
) -> List[ObjectAccess]:
    accesses = model_bulk_get_or_create(ObjectAccess, [
        ({'user_id': user_id, 'object_id': id, 'content_type': ct}, {
            'permission_ids': permission_ids or [],
            'group_ids': group_ids or [],
        })
        for ct, id in object_definitions
    ])

    for access in accesses:
        access.permission_ids = committer(access.permission_ids, permission_ids)
        access.group_ids = committer(access.group_ids, group_ids)

    # TODO: Should remove access models here if there is an empty
    # `permission_ids` and `groups_ids` attributes.

    accesses = gatherer.gather_object_accesses(accesses, update=False)

    ObjectAccess.objects.bulk_update(accesses, fields=(
        'group_ids', 'permission_ids', 'gathered_permission_ids'
    ))

    return accesses


def remove_access_raw(
    user_id: int,
    object_definitions: Sequence[ObjectDefinition],
):
    q = models.Q()

    for ct, id in object_definitions:
        q |= models.Q(object_id=id, content_type=ct)

    ObjectAccess.objects.filter(q, user_id=user_id).delete()


def ADD_COMMITTER(has: Sequence[int], add: Optional[Sequence[int]] = None) -> List[int]:
    return list(has if add is None else upset(has, add))


def add_access_raw(
    user_id: int,
    object_definitions: Sequence[ObjectDefinition],
    permission_ids: Optional[Sequence[int]] = None,
    group_ids: Optional[Sequence[int]] = None,
) -> List[ObjectAccess]:
    if not any(_should(permission_ids, group_ids)):
        return []

    return change_access_raw(
        user_id, object_definitions, ADD_COMMITTER,
        permission_ids=permission_ids, group_ids=group_ids,
    )


def WITHDRAW_COMMITTER(has: Sequence[int], withdraw: Optional[Sequence[int]] = None) -> List[int]:
    return list(has if withdraw is None else (set(has) - set(withdraw)))


def withdraw_access_raw(
    user_id: int,
    object_definitions: Sequence[ObjectDefinition],
    permission_ids: Optional[Sequence[int]] = None,
    group_ids: Optional[Sequence[int]] = None,
) -> List[ObjectAccess]:
    if not any(_should(permission_ids, group_ids)):
        return []

    return change_access_raw(
        user_id, object_definitions, WITHDRAW_COMMITTER,
        permission_ids=permission_ids, group_ids=group_ids,
    )


def SET_COMMITTER(has: Sequence[int], to_set: Optional[Sequence[int]] = None) -> List[int]:
    return list(has if to_set is None else to_set)


def set_access_raw(
    user_id: int,
    object_definitions: Sequence[ObjectDefinition],
    permission_ids: Optional[Sequence[int]] = None,
    group_ids: Optional[Sequence[int]] = None,
) -> List[ObjectAccess]:
    if (
        permission_ids is not None
        and
        group_ids is not None
        and
        len(permission_ids) == 0
        and
        len(group_ids) == 0
    ):
        remove_access_raw(user_id, object_definitions)

        return []

    return change_access_raw(
        user_id, object_definitions, SET_COMMITTER,
        permission_ids=permission_ids, group_ids=group_ids,
    )


def add_access(
    user_id: int,
    objects: Sequence[models.Model],
    permission_ids: Optional[Sequence[int]] = None,
    group_ids: Optional[Sequence[int]] = None,
) -> List[ObjectAccess]:
    return add_access_raw(
        user_id, _obj_to_raw(objects),
        permission_ids=permission_ids, group_ids=group_ids,
    )


def withdraw_access(
    user_id: int,
    objects: Sequence[models.Model],
    permission_ids: Optional[Sequence[int]] = None,
    group_ids: Optional[Sequence[int]] = None,
) -> List[ObjectAccess]:
    return withdraw_access_raw(
        user_id, _obj_to_raw(objects),
        permission_ids=permission_ids, group_ids=group_ids,
    )


def remove_access(user_id: int, objects: Sequence[models.Model]):
    return remove_access_raw(user_id, _obj_to_raw(objects))


def set_access(
    user_id: int,
    objects: Sequence[models.Model],
    permission_ids: Optional[Sequence[int]] = None,
    group_ids: Optional[Sequence[int]] = None,
) -> List[ObjectAccess]:
    return set_access_raw(
        user_id, _obj_to_raw(objects),
        permission_ids=permission_ids, group_ids=group_ids,
    )



from django.contrib import admin
from django.contrib.contenttypes.models import ContentType

from pxd_combinable_groups.utils import get_permission_model

from .services import gatherer
from .models import ObjectAccess
from .conf import settings


Permission = get_permission_model()


@admin.register(ObjectAccess)
class ObjectAccessAdmin(admin.ModelAdmin):
    list_display = 'id', 'content_type', 'object_id', 'user'
    list_select_related = 'user', 'content_type',
    autocomplete_fields = 'user',
    list_filter = 'content_type',
    search_fields = (
        'content_type__app_label', 'content_type__model',
        'user__username', 'object_id',
    )

    def save_model(self, request, obj, form, change):
        result = super().save_model(request, obj, form, change)

        gatherer.gather_object_accesses((obj,))

        return result


class OnlyViewMixin:
    has_delete_permission = lambda *a: False
    has_change_permission = lambda *a: False
    has_add_permission = lambda *a: False


if settings.ADMIN_REGISTER_PERMISSION and not admin.site.is_registered(Permission):
    @admin.register(Permission)
    class PermissionAdmin(OnlyViewMixin, admin.ModelAdmin):
        list_display = '__str__', 'id', 'codename', 'content_type',
        list_select_related = 'content_type',
        list_filter = 'content_type',
        search_fields = (
            'content_type__app_label', 'content_type__model',
            'codename',
        )


if settings.ADMIN_REGISTER_CONTENT_TYPE and not admin.site.is_registered(ContentType):
    @admin.register(ContentType)
    class ContentTypeAdmin(OnlyViewMixin, admin.ModelAdmin):
        list_display = '__str__', 'id', 'app_label', 'model',
        list_filter = 'app_label',



from typing import TYPE_CHECKING, Callable, Optional, Sequence, Tuple, Type, TypeVar
from functools import wraps, cached_property
from django.db import models
from django.contrib.contenttypes.models import ContentType

from pxd_combinable_groups.services import permissions_collector
from pxd_combinable_groups.utils import resolve_on_cached_object

from ..models import ObjectAccess
from ..exceptions import CheckError
from ..utils import make_random_code

if TYPE_CHECKING:
    from django.contrib.auth.models import AbstractBaseUser as User


CHECKER_ANNOTATION_FIELD_TEMPLATE = '_permissions_{name}_exists'


def require_permissions(permission_ids: Sequence[int] = None):
    if permission_ids is None or len(permission_ids) == 0:
        raise CheckError('At least one permission required to compare against.')


def wrap_require_permissions(fn):
    @wraps(fn)
    def wrapped(*a, permission_ids: Optional[Sequence[int]] = None, **kw):
        require_permissions(permission_ids)

        return fn(*a, permission_ids, **kw)

    return wrapped


class CheckCMP:
    def cmp_Q(self, field: str, permission_ids: Sequence[int] = None):
        raise NotImplementedError()

    def cmp(self, ids: Sequence[int], permission_ids: Sequence[int] = None):
        raise NotImplementedError()


class RawCheckCMP(CheckCMP):
    def __init__(self, cmp_Q: Callable, cmp: Callable):
        self.cmp_Q = cmp_Q
        self.cmp = cmp


ALL_CMP = RawCheckCMP(
    wrap_require_permissions(
        lambda field, pids: models.Q(**{f'{field}__contains': pids})
    ),
    wrap_require_permissions(
        lambda ids, pids: len(set(ids) & set(pids)) == len(pids)
    )
)
ANY_CMP = RawCheckCMP(
    wrap_require_permissions(
        lambda field, pids: models.Q(**{f'{field}__overlap': pids})
    ),
    wrap_require_permissions(
        lambda ids, pids: len(set(ids) & set(pids)) > 0
    )
)


# TODO: Remove this.
# def resolve_reverse(path: Sequence[str], model: Type[models.Model]):
#     raise NotImplementedError()
#     field_name = path[0]
#     field = model._meta.get_field(field_name)
#     rel = field.remote_field

#     from pprint import pprint; pprint(field.__dict__)
#     if isinstance(rel, models.ManyToManyRel):
#         # prepath = resolve_reverse(
#         #     [field.m2m_reverse_field_name()] + path[1:],
#         #     rel.through
#         # )
#         # next_field = rel.through._meta.get_field()

#         # print(field)
#         # print(rel)
#         # pprint(rel.__dict__)
#         print('r', field.m2m_reverse_name())
#         # print(field.m2m_target_name())
#         print(field.m2m_reverse_field_name())
#         # print(field.m2m_field_name())
#         # print(field.m2m_column_name())
#         # pprint(next_field.__dict__)
#     if isinstance(rel, models.ManyToOneRel):
#         # pprint(rel.__dict__)
#         pass
#     else:
#         print('else', rel)

#     # for field_name in path:
#     #     field = model._meta.get_field(field_name)

#     #     print(field)

#     return path, model


class Checker:
    path: str
    model: Type[models.Model]
    root_model: Type[models.Model]
    cmp: CheckCMP = ALL_CMP
    should_check_global: bool = True

    def __init__(
        self,
        model: Type[models.Model],
        path: str,
        root_model: Optional[Type[models.Model]] = None,
        cmp: Optional[CheckCMP] = None,
        should_check_global: Optional[bool] = None,
    ):
        self.path = path
        self.model = model
        # TODO: Remove this.
        # self.reverse_path, founded_model = self._resolve_reverse(path, model)
        self.root_model = root_model
        self.cmp = cmp if cmp is not None else self.cmp
        self.should_check_global = (
            should_check_global if should_check_global is not None
            else self.should_check_global
        )

        assert self.root_model is not None, (
            'Cant resolve root model provide it manually please.'
        )

    @cached_property
    def content_type(self):
        return ContentType.objects.get_for_model(self.root_model)

    # TODO: Remove this.
    # def _resolve_reverse(self, path: str, model: Type[models.Model]):
    #     resolve_reverse(path.split('__'), model)
    #     return path, model

    def _make_annotation_field_name(self, name: str) -> str:
        return CHECKER_ANNOTATION_FIELD_TEMPLATE.format(name=name)

    def _prepare_access_query(
        self,
        user: 'User',
        permission_ids: Optional[Sequence[int]] = None,
        cmp: Optional[CheckCMP] = None,
    ) -> models.QuerySet:
        # TODO: Separate query somewhere.
        return ObjectAccess.objects.filter(
            (cmp or self.cmp).cmp_Q(
                'gathered_permission_ids',
                permission_ids=permission_ids,
            ),
            user_id=user.id, content_type=self.content_type,
        )

    def has_global_access(
        self,
        user: 'User',
        permission_ids: Optional[Sequence[int]] = None,
        cmp: Optional[CheckCMP] = None,
    ) -> bool:
        if not self.should_check_global:
            return False

        ids = permissions_collector.for_user(user)

        return (cmp or self.cmp).cmp(ids, permission_ids=permission_ids)

    def with_annotation(
        self,
        queryset: models.QuerySet,
        user: 'User',
        permission_ids: Optional[Sequence[int]] = None,
        cmp: Optional[CheckCMP] = None,
        annotation_field_name: Optional[str] = None,
    ) -> Tuple[str, models.QuerySet]:
        annotation_field_name = (
            annotation_field_name
            or
            self._make_annotation_field_name(make_random_code(6))
        )

        if self.has_global_access(user, permission_ids=permission_ids, cmp=cmp):
            return annotation_field_name, queryset.annotate(**{
                annotation_field_name: models.Value(True)
            })

        subquery = self._prepare_access_query(
            user, permission_ids=permission_ids, cmp=cmp,
        ).filter(
            object_id=models.OuterRef(self.path)
        )

        return annotation_field_name, queryset.annotate(**{
            annotation_field_name: models.Exists(subquery)
        })

    def for_object(
        self,
        obj: models.Model,
        user: 'User',
        permission_ids: Sequence[int],
        cmp: Optional[CheckCMP] = None,
    ) -> bool:
        if self.has_global_access(user, permission_ids=permission_ids, cmp=cmp):
            return True

        # FIXME: An ugly query here. try to change it somehow.
        permission_object = resolve_on_cached_object(
            user, f'_object_access_{self.content_type.pk}_{obj.pk}',
            lambda *a: (
                self._prepare_access_query(
                    user, permission_ids=permission_ids, cmp=cmp
                )
                .filter(object_id__in=(
                    self.model.objects.filter(pk=obj.pk).values(self.path)
                ))
                .first()
            )
        )

        if permission_object is None:
            return False

        return (cmp or self.cmp).cmp(
            permission_object.gathered_permission_ids,
            permission_ids=permission_ids,
        )

    def for_queryset(
        self,
        queryset: models.QuerySet,
        user: 'User',
        permission_ids: Optional[Sequence[int]] = None,
        cmp: Optional[CheckCMP] = None,
        annotation_field_name: Optional[str] = None,
    ) -> models.QuerySet:
        name, q = self.with_annotation(
            queryset, user,
            permission_ids=permission_ids, cmp=cmp,
            annotation_field_name=annotation_field_name,
        )

        return q.filter(**{name: True})



# Generated by Django 3.2.14 on 2022-08-05 06:45

from django.conf import settings
import django.contrib.postgres.fields
import django.contrib.postgres.indexes
from django.contrib.postgres.operations import CreateExtension
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        CreateExtension('intarray'),

        migrations.CreateModel(
            name='ObjectAccess',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField(verbose_name='Object: Identifier')),
                ('gathered_permission_ids', django.contrib.postgres.fields.ArrayField(base_field=models.IntegerField(), blank=True, default=list, size=None, verbose_name='Gathered permissions')),
                ('group_ids', django.contrib.postgres.fields.ArrayField(base_field=models.IntegerField(), blank=True, default=list, size=None, verbose_name='Groups')),
                ('permission_ids', django.contrib.postgres.fields.ArrayField(base_field=models.IntegerField(), blank=True, default=list, size=None, verbose_name='Permissions')),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype', verbose_name='Object: Content type')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User')),
            ],
            options={
                'verbose_name': 'Object access',
                'verbose_name_plural': 'Object accesses',
            },
        ),
        migrations.AddIndex(
            model_name='objectaccess',
            index=models.Index(fields=['content_type', 'object_id'], name='pxd_legate__content_1fe8f8_idx'),
        ),
        migrations.AddIndex(
            model_name='objectaccess',
            index=django.contrib.postgres.indexes.GinIndex(fields=['gathered_permission_ids'], name='pxd_legate_objectaccess_gth', opclasses=('gin__int_ops',)),
        ),
        migrations.AddIndex(
            model_name='objectaccess',
            index=django.contrib.postgres.indexes.GinIndex(fields=['group_ids'], name='pxd_legate_objectaccess_groups', opclasses=('gin__int_ops',)),
        ),
        migrations.AddIndex(
            model_name='objectaccess',
            index=django.contrib.postgres.indexes.GinIndex(fields=['permission_ids'], name='pxd_legate_objectaccess_perms', opclasses=('gin__int_ops',)),
        ),
        migrations.AlterUniqueTogether(
            name='objectaccess',
            unique_together={('user', 'content_type', 'object_id')},
        ),
    ]



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.fields import ArrayField
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import pgettext_lazy
from django.db import models


class GatheredPermissionsBase(models.Model):
    class Meta:
        abstract = True

    content_type = models.ForeignKey(
        ContentType, on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('pxd_legate', 'Object: Content type'),
    )
    object_id = models.PositiveIntegerField(
        verbose_name=pgettext_lazy('pxd_legate', 'Object: Identifier'),
    )
    content_object = GenericForeignKey('content_type', 'object_id')

    gathered_permission_ids = ArrayField(
        models.IntegerField(), blank=True, default=list,
        verbose_name=pgettext_lazy('pxd_legate', 'Gathered permissions'),
    )


def gathered_permission_indexes(name: str = '%(app_label)s_%(class)s'):
    return [
        models.Index(fields=['content_type', 'object_id']),
        GinIndex(
            name=name + '_gth',
            fields=['gathered_permission_ids'], opclasses=('gin__int_ops',)
        )
    ]



from typing import Dict, List, Sequence
from django.utils.translation import pgettext_lazy
from django.db import models

from .utils import get_group_model


Group = get_group_model()


class GroupToGroupConnectionQuerySet(models.QuerySet):
    pass


class GroupToGroupConnection(models.Model):
    objects = GroupToGroupConnectionQuerySet.as_manager()

    class Meta:
        verbose_name = pgettext_lazy(
            'pxd_legate', 'Group to group connection'
        )
        verbose_name_plural = pgettext_lazy(
            'pxd_legate', 'Group to group connection'
        )
        unique_together = (
            ('grouper', 'source'),
        )

    grouper = models.ForeignKey(
        Group, related_name='source_connections', on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('pxd_legate', 'Grouper group'),
    )
    source = models.ForeignKey(
        Group, related_name='grouper_connections', on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('pxd_legate', 'Source group'),
    )

    def __str__(self):
        return f'#{self.grouper_id} -> #{self.source_id}'



from typing import List, Sequence

from pxd_combinable_groups.services import permissions_collector

from ..models import ObjectAccess
from ..utils import upset


def gather_object_accesses(
    accesses: Sequence[ObjectAccess],
    update: bool = True,
) -> List[ObjectAccess]:
    collected = permissions_collector.collect_sets(
        access.group_ids
        for access in accesses
    )
    accesses = accesses if isinstance(accesses, list) else list(accesses)

    for i, permissions in enumerate(collected):
        accesses[i].gathered_permission_ids = list(upset(
            permissions, accesses[i].permission_ids
        ))

    if update:
        ObjectAccess.objects.bulk_update(accesses, fields=(
            'gathered_permission_ids',
        ))

    return accesses



from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.postgres.fields import ArrayField
from django.contrib.contenttypes.models import ContentType
from django.utils.translation import pgettext_lazy
from django.conf import settings
from django.db import models
from django.contrib.postgres.indexes import GinIndex

from .gathered import GatheredPermissionsBase, gathered_permission_indexes


class ObjectAccess(GatheredPermissionsBase):
    class Meta:
        verbose_name = pgettext_lazy('pxd_legate', 'Object access')
        verbose_name_plural = pgettext_lazy('pxd_legate', 'Object accesses')
        unique_together = (
            ('user', 'content_type', 'object_id'),
        )
        indexes = gathered_permission_indexes() + [
            GinIndex(
                name='%(app_label)s_%(class)s_groups',
                fields=['group_ids'], opclasses=('gin__int_ops',)
            ),
            GinIndex(
                name='%(app_label)s_%(class)s_perms',
                fields=['permission_ids'], opclasses=('gin__int_ops',)
            ),
        ]

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('pxd_legate', 'User'),
    )
    group_ids = ArrayField(
        models.IntegerField(), blank=True, default=list,
        verbose_name=pgettext_lazy('pxd_legate', 'Groups'),
    )
    permission_ids = ArrayField(
        models.IntegerField(), blank=True, default=list,
        verbose_name=pgettext_lazy('pxd_legate', 'Permissions'),
    )



import string
import random
from functools import lru_cache
from typing import Sequence
from django.utils.module_loading import import_string


ALPHABET = string.ascii_letters.split()


@lru_cache
def cached_import_string(path: str):
    return import_string(path)


def upset(current, add):
    return (
        (current if isinstance(current, set) else set(current))
        |
        (add if isinstance(add, set) else set(add))
    )


def make_random_code(symbols: int, alphabet: Sequence[str] = ALPHABET) -> str:
    return ''.join(random.sample(alphabet * symbols, symbols))



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = ('LegateConfig',)


class LegateConfig(AppConfig):
    name = 'pxd_legate'
    verbose_name = pgettext_lazy('pxd_legate', 'Legate')
